/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.valueprovider;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;

/**
 * Base class for {@link IAnnotationValueProvider}s that manages {@link EStructuralFeature}s
 * contained by a {@link IAnnotatedSpecification}.
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public abstract class EStructuralFeatureValueProviderBase<T extends IAnnotatedSpecification>
		extends ValueProviderBase<T> {

	/**
	 * Default key used in case this {@link IAnnotationValueProvider} is used to manage a single
	 * {@link EStructuralFeature}, only.
	 */
	final static String DEFAULT_KEY = null;

	/**
	 * Map: Instance key -> {@link EStructuralFeatureDescriptor}s of the
	 * {@link IAnnotatedSpecification} managed by
	 * this {@link IAnnotationValueProvider}.
	 */
	final protected Map<String, EStructuralFeatureDescriptor> structuralFeatureDescriptorMap;

	/**
	 * Flag if multiple {@link EStructuralFeature}s are managed by this
	 * {@link IAnnotationValueProvider}.
	 */
	boolean handlesMultipleEstructuralFeatures;

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented by a single {@link EStructuralFeature}, and that uses the
	 * default constructor to construct a new value for the structural feature.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EFactory annotationFactory, EStructuralFeature structuralFeature,
			EFactory structuralFeatureFactory) {
		this(annotatedSpecificationEClass, annotationFactory, structuralFeature,
				structuralFeatureFactory, null);
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented by a single {@link EStructuralFeature}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EFactory annotationFactory, EStructuralFeature structuralFeature,
			EFactory structuralFeatureFactory, final String structuralFeatureDefaultValue) {

		super(annotatedSpecificationEClass, annotationFactory);

		structuralFeatureDescriptorMap = new HashMap<String, EStructuralFeatureDescriptor>();
		structuralFeatureDescriptorMap.put(DEFAULT_KEY, new EStructuralFeatureDescriptor(
				structuralFeature, structuralFeatureFactory, structuralFeatureDefaultValue));

		handlesMultipleEstructuralFeatures = false;
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented by a number of {@link EStructuralFeature}s contained in a single
	 * EClass, and that uses the default constructor to construct a new value for the structural
	 * features.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EFactory annotationFactory,
			Map<String, EStructuralFeatureDescriptor> structuralFeatureDescriptorMap) {

		super(annotatedSpecificationEClass, annotationFactory);

		this.structuralFeatureDescriptorMap = structuralFeatureDescriptorMap;

		handlesMultipleEstructuralFeatures = true;
	}

	/**
	 * Returns the {@link EStructuralFeatureDescriptor} for a given {@code instanceKey}, or throws
	 * an exception, if the key is unknown.
	 * 
	 */
	protected EStructuralFeatureDescriptor getEStructuralFeatureDescriptor(String instanceKey)
			throws Exception {
		EStructuralFeatureDescriptor fd = structuralFeatureDescriptorMap.get(instanceKey);
		if(fd == null) {
			throw new Exception("Instance key " + instanceKey +
					" is unknown for annotation provider " +
					annotatedSpecificationEClass.getInstanceClassName() + ".");
		}

		return fd;
	}

	/**
	 * Sets a value for a {@link IAnnotatedSpecification} from a {@link String} representation of
	 * the input.
	 */
	private <V> void
			setAnnotationValueFromString(String value, T specification, String instanceKey)
					throws Exception {

		EStructuralFeatureDescriptor fd = getEStructuralFeatureDescriptor(instanceKey);

		specification.eSet(
				fd.getStructuralFeature(),
				fd.getStructuralFeatureFactory().createFromString(
						((EAttribute)fd.getStructuralFeature()).getEAttributeType(), value));
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> void setAnnotationValue(U value, T specification, String instanceKey)
			throws Exception {
		if(value instanceof String) {
			// In case, the input choice is presented in dynamically managed combo box (i.e., where
			// new value choices can be entered by the user, the empty String indicates that the
			// current value should be deleted, and a neighboring value from the set should be used.
			// In case the last occurrence of a value choice has been removed, it will be no longer
			// be offered in the combo box.
			if(((String)value).isEmpty()) {
				String rval = getAnnotationValueAndUpdateInputChoice((String)value, specification);
				if(rval != null) {
					value = (U)rval;
				}
			}

			setAnnotationValueFromString((String)value, specification, instanceKey);
		} else {
			specification.eSet(getEStructuralFeatureDescriptor(instanceKey).getStructuralFeature(),
					value);
		}
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification) throws Exception {
		setAnnotationValue(value, specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification, String instanceKey)
			throws Exception {
		// See comment in setAnnotationValue(U value, T specification)
		if((value == null || value.isEmpty())) {
			String rval = getAnnotationValueAndUpdateInputChoice(value, specification);
			if(rval != null) {
				value = rval;
			}
		}

		EStructuralFeature structuralFeature =
				getEStructuralFeatureDescriptor(instanceKey).getStructuralFeature();

		if(structuralFeature instanceof EAttribute) {
			setAnnotationValueFromString(value, specification, instanceKey);
		} else {
			throw new Exception(
					"setAnnotationValue(String, T) is not supported / has not been implemented the annotation type " +
							structuralFeature.getEType().getName() +
							"(specification class: " +
							specification.getClass() + ").");
		}
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification) throws Exception {
		setAnnotationValue(value, specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception {
		return (U)specification.eGet(getEStructuralFeatureDescriptor(instanceKey)
				.getStructuralFeature());
	}

	/** {@inheritDoc} */
	@Override
	public <U> U getAnnotationValue(T specification) throws Exception {
		return getAnnotationValue(specification, DEFAULT_KEY);
	}

	/**
	 * Creates a new instance of the {@link EStructuralFeature} (use to populate newly created
	 * annotation instances created in
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}.
	 */
	protected Object createStructuralFeatureInstance(String instanceKey) throws Exception {
		EStructuralFeatureDescriptor fd = getEStructuralFeatureDescriptor(instanceKey);

		if(fd.getStructuralFeature() instanceof EAttribute) {

			EDataType attributeEDataType =
					((EAttribute)fd.getStructuralFeature()).getEAttributeType();

			if(fd.getStructuralFeatureDefaultValue() != null) {
				// Use default value if it has been specified
				return fd.structuralFeatureFactory.createFromString(attributeEDataType,
						fd.getStructuralFeatureDefaultValue());
			}

			// Return data types default value, if it is non-null.
			Object rval = attributeEDataType.getDefaultValue();

			if(rval != null) {
				return rval;
			}

			try {
				// Try to construct attribute from empty string
				rval = fd.getStructuralFeatureFactory().createFromString(attributeEDataType, "");
			} catch(NumberFormatException e) {
				// Last try: initialize numbers with 0 (throws NumberFormatException if not
				// successful).
				rval = fd.getStructuralFeatureFactory().createFromString(attributeEDataType, "0");
			}

			return rval;
		} else if(fd.getStructuralFeature() instanceof EReference) {
			EReference reference = (EReference)fd.getStructuralFeature();
			if(reference.isContainment()) {
				// EClasses attached to contained EReferences are created lazily by the
				// corresponding generated code on the first access (i.e., also during get / "read")
				return null;
			}
			return fd.getStructuralFeatureFactory().create(reference.getEReferenceType());
		} else {
			throw new Exception(
					"createStructuralFeatureInstance() is not supported / has not been implemented the annotation type " +
							fd.getStructuralFeature().getName() + ".");
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void decorateAnnotationSpecification(T specification, String instanceKey)
			throws Exception {

		// Create and set structural feature implementing the annotation
		Object structuralFeatureVal = createStructuralFeatureInstance(instanceKey);
		if(structuralFeatureVal != null) {
			specification.eSet(getEStructuralFeatureDescriptor(instanceKey).getStructuralFeature(),
					structuralFeatureVal);
		}
	}

	/** {@inheritDoc} */
	@Override
	public Collection<String> getInstanceKeys(T specification) {
		if(handlesMultipleEstructuralFeatures) {
			return structuralFeatureDescriptorMap.keySet();
		}
		return super.getInstanceKeys(specification);
	}
}
